/* [vphasor~]
 * 
 * [phasor~] with a sub-sample accurate phase reset
 * 
 * by Mike Moser-Booth
 * with code borrowed from Miller Puckette, of course ;-)
 */

#include "m_pd.h"

/* All this jazz is needed to implement the Hoeldrich
 * version of a phasor. From what I understand, it essentially
 * does wrapping by forcing the integer part of a union of
 * a double and a two-int32 array (tabfudge) by resetting the
 * MSB of the double to UNITBIT32. Which one it is depends on
 * endianess. */

#define UNITBIT32 1572864.  /* 3*2^19; bit 32 has place value 1 */


#if defined(__FreeBSD__) || defined(__APPLE__) || defined(__FreeBSD_kernel__) \
    || defined(__OpenBSD__)
#include <machine/endian.h>
#endif

#if defined(__linux__) || defined(__CYGWIN__) || defined(__GNU__) || \
    defined(ANDROID)
#include <endian.h>
#endif

#ifdef __MINGW32__
#include <sys/param.h>
#endif

#ifdef _MSC_VER
/* _MSVC lacks BYTE_ORDER and LITTLE_ENDIAN */
#define LITTLE_ENDIAN 0x0001
#define BYTE_ORDER LITTLE_ENDIAN
#endif

#if !defined(BYTE_ORDER) || !defined(LITTLE_ENDIAN)
#error No byte order defined
#endif

#if BYTE_ORDER == LITTLE_ENDIAN
# define HIOFFSET 1
# define LOWOFFSET 0
#else
# define HIOFFSET 0    /* word offset to find MSB */
# define LOWOFFSET 1    /* word offset to find LSB */
#endif

union tabfudge
{
    double tf_d;
    int32_t tf_i[2];
};

/* --------------------- vphasor~ --------------------*/
static t_class *vphasor_tilde_class;

typedef struct _vreset {
  double	s_resettime;	/* time to reset phase */
  double	s_resetphase;	/* value to reset phase at */
  struct _vreset *s_next;	/* link to next in list */
} t_vreset;

typedef struct _vphasor_tilde {
  t_object	x_obj;
  float		x_deffreq;	/* default frequency */
  
  double	x_phase;	/* current phase */
  float		x_oneoversr;	/* 1/SR, for converting frequency from 
				 * cycles/second to cycles/sample */
  double	x_samplesperms;
  double	x_mspersample;
  double	x_blocksizems;	/* blocksize in ms */
  double	x_blocktime;	/* logical time of start of block */
 
  t_vreset	*x_list;	/* list of phase reset values and times */
} t_vphasor_tilde;

static void *vphasor_tilde_new(t_floatarg deffreq)
{
  t_vphasor_tilde *x = (t_vphasor_tilde *)pd_new(vphasor_tilde_class);
  
  x->x_deffreq = deffreq;
  x->x_phase = 0.;
  x->x_oneoversr = 0.;
  x->x_samplesperms = 0.;
  x->x_mspersample = 0.;
  x->x_blocksizems = 0.;
  x->x_blocktime = 0.;
  x->x_list = 0;
  
  inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("phase"));
  outlet_new(&x->x_obj, &s_signal);
  
  return (void *)x;
}

static t_int *vphasor_tilde_perform(t_int *w)
{
  t_vphasor_tilde *x = (t_vphasor_tilde *)(w[1]);
  t_sample *freqin = (t_sample *)(w[2]);
  t_sample *phaseout = (t_sample *)(w[3]);
  int blocksize = (int)(w[4]);
  
  double mspersample = x->x_mspersample;
  double samplesperms = x->x_samplesperms;
  double blocktimeelapsed = 0.;
  float oneoversr = x->x_oneoversr;
  double dphase = x->x_phase + (double)UNITBIT32;
  union tabfudge tf;
  int normhipart;
  
  t_vreset *s1 = x->x_list, *s2;
  
  /* get the integer part */
  tf.tf_d = UNITBIT32;
  normhipart = tf.tf_i[HIOFFSET];
  
  while (blocksize--)
  {
    /* This is where the magic happens */
    /* if there's no phase to reset, don't bother checking */
    while (s1)
    {
      /* This is in a while-loop because it may be possible
       * that more than one phase reset is scheduled within
       * the same sample length.
       * Once we find a resettime that hasn't happened yet,
       * we can stop checking. */
      if (s1->s_resettime > blocktimeelapsed)
	break;
      
      /* get the sub-sample length since the phase was reset to
       * the current sample */
      double sampfrac = (blocktimeelapsed - s1->s_resettime) * samplesperms;
      /* take the resetphase and add the sub-sample length of
       * accumulated phase to it, and make it the current phase */
      dphase = s1->s_resetphase + ((*freqin) * oneoversr * sampfrac) + (double)UNITBIT32;
      /* Once the phase is reset, we can free it from the list
       * and memory */
      s2 = s1->s_next;
      freebytes(s1, sizeof(*s1));
      s1 = s2;
    }
    
    x->x_list = s1;
    
    tf.tf_d = dphase;  
    /* wrap phase so that 0 <= phase < 1
     * by forcing the integer part of dphase without affecting
     * the fractional part (which is the actual phase we want),
     * then subtracting the integer part at the output.
     * ...if I understand that correctly, that is */
    tf.tf_i[HIOFFSET] = normhipart;
    /* We output the current sample before updating the phase,
     * because if we start at zero we better output zero. But we
     * still have to read the inlet beforehand because the inlet
     * and outlet may share the same buffer, so we need to read
     * it before we write over it. */
    dphase += *freqin++ * oneoversr;
    *phaseout++ = tf.tf_d - UNITBIT32;
    tf.tf_d = dphase;
    
    blocktimeelapsed += mspersample;
  }
  
  /* The loop above considers the time from 0 <= t <= n-1
   * It doesn't take care of n-1 to n (which is 0 in the next
   * block. So if the phase resettime is inbetween n-1 and n,
   * it will get skipped. So we pull the resettime
   * from the blocksize in ms so that it will get checked at the
   * next block, if the resettime still exists. This should give
   * a negative time value less than one sample long, so it will 
   * be processed on the first sample of the next block. */
  s1 = x->x_list;
  while (s1)
  {
    s1->s_resettime = s1->s_resettime - x->x_blocksizems;
    s1 = s1->s_next;
  }
//  if (x->x_resettime != -1.)
//    x->x_resettime = x->x_resettime - x->x_blocksizems;
  
  x->x_blocktime = clock_getlogicaltime();
  tf.tf_i[HIOFFSET] = normhipart;
  x->x_phase = tf.tf_d - UNITBIT32;
  return (w+5);
}

static void vphasor_tilde_dsp(t_vphasor_tilde *x, t_signal **sp)
{
  x->x_oneoversr = 1. / sp[0]->s_sr;
  x->x_samplesperms = ((double)(sp[0]->s_sr)) / 1000;
  x->x_mspersample = ((double)1000) / sp[0]->s_sr;
  x->x_blocksizems = sp[0]->s_n * x->x_mspersample;
  
  dsp_add(vphasor_tilde_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

static void vphasor_tilde_freelist(t_vphasor_tilde *x)
{
  /* Go through the list of stored resets and free
   * the memory they take up */
  t_vreset *s1, *s2;
  for(s1 = x->x_list; s1; s1 = s2)
  {
    s2 = s1->s_next;
    freebytes(s1, sizeof(*s1));
  }
  x->x_list = 0;
}

static void vphasor_tilde_phase(t_vphasor_tilde *x, t_floatarg phase)
{
  double elapsedtime = clock_gettimesince(x->x_blocktime);
  t_vreset *snew, *s1, *s2;
  
  if (PD_BIGORSMALL(phase))
    phase = 0.;
  
  if(elapsedtime > x->x_blocksizems)
  {
    /* For now we'll just assume that if the elapsed time is
     * greater that than the block size that dsp has
     * been turned off, so we'll just immediately reset
     * the phase when the block starts 
     * TODO: Is this a safe assumption? See if there
     * is a better solution. Maybe deal with it in
     * vphasor_tilde_dsp()? */
    x->x_phase = phase;
    /* clear out any stored phase resets */
    vphasor_tilde_freelist(x);
  }
  else
  {
    snew = (t_vreset *)getbytes(sizeof(*snew));
    if(!x->x_list)
    {
       x->x_list = snew;
    }
    else
    {
      /* find end of list */
      for(s1 = x->x_list; (s1->s_next); s1 = s2)
        s2 = s1->s_next;
      s1->s_next = snew;
    }
    
    snew->s_resettime = elapsedtime;
    snew->s_resetphase = phase;
    snew->s_next = 0;
  }
}

void vphasor_tilde_setup(void)
{
  vphasor_tilde_class = class_new(gensym("vphasor~"),
				  (t_newmethod)vphasor_tilde_new,
				  (t_method)vphasor_tilde_freelist,
				  sizeof(t_vphasor_tilde),
				  CLASS_DEFAULT,
				  A_DEFFLOAT,
				  0);
  
  CLASS_MAINSIGNALIN(vphasor_tilde_class, t_vphasor_tilde, x_deffreq);
  class_addmethod(vphasor_tilde_class, (t_method)vphasor_tilde_dsp,
		  gensym("dsp"), 0);
  class_addmethod(vphasor_tilde_class, (t_method)vphasor_tilde_phase,
		  gensym("phase"), A_FLOAT, 0);
}